
# THIS FILE IS GENERATED FROM OASYS SETUP.PY
short_version = '1.2.2'
version = '1.2.2'
full_version = '1.2.2'
git_revision = '397d3e28af6bfab5ebec85cdd2874bee8de284dc'
release = True

if not release:
    version = full_version
    short_version += ".dev"
